#include <linux/module.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/clk.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <asm/arch/gpio.h>
#include <asm/arch/at91_tc.h>
#include <asm/arch/at91_adc.h>
#include <asm/uaccess.h>


static dev_t at91adc_devno;
static struct cdev at91adc_cdev;
static struct file_operations at91adc_fops;

static unsigned char *adc_buf, pref;
static dma_addr_t rx_dma;
static unsigned int MCK, curr_srate, state, srate, hsize, threshold;
static void __iomem *tc0_base;
static void __iomem *adc_base;
static struct clk *adc_clk;
static struct clk *tc0_clk;


#define PCF8833 128
#define START_THRESHOLD 10

#define MAX_BUF_LEN 1024
#define MIN_SRATE 50
#define MAX_SRATE 312000
#define SINGLE_BYTE 0
#define MULTI_BYTE 1

#define at91_adc_read(reg)		__raw_readl(adc_base + (reg))
#define at91_adc_write(reg, val)	__raw_writel((val), adc_base + (reg))

#define at91_tc_read(reg)		__raw_readl(tc0_base + (reg))
#define at91_tc_write(reg, val)	__raw_writel((val), tc0_base + (reg))


static unsigned int align_rate(unsigned int rate);
{
    for( ; rate >= MIN_SRATE; rate--)
		if((MCK/32) == ((MCK/32) / rate * rate))
			break;                                                              

	return rate;
}

static irqreturn_t at91adc_isr(int irq, void *dev_id)
{
	/* reset dma pointer  */
	at91_adc_write(ATMEL_PDC_RPR, rx_dma);

	switch(state) {
	case MULTI_BYTE:
		at91_adc_write(ATMEL_PDC_RCR, 1);
		state = SINGLE_BYTE;
		pref = 255;
		break;
	case SINGLE_BYTE:
		if((adc_buf[0] >= threshold) && (pref < threshold)) {
			at91_adc_write(ATMEL_PDC_RCR, hsize);
			state = MULTI_BYTE;
		} else {
			at91_adc_write(ATMEL_PDC_RCR, 1);
			pref = adc_buf[0];
		}
		break;
	}

	if(srate != curr_srate) {
		srate = align_rate(srate);
		if(srate >= MIN_SRATE) { 
			curr_srate = srate;
			at91_tc_write(AT91_TC_RC, (MCK/32) / curr_srate);
			/* duty 50 % */
			at91_tc_write(AT91_TC_RA, (MCK/32) / (curr_srate * 2));
			pref = 255;
		}
		else 
			srate = curr_srate;
	}

	return IRQ_HANDLED;
}

static int __init at91adc_init(void)
{
	int ret;
	unsigned int  ADCCLK, prescal, shtim;
	struct clk *mck_clk;

	adc_clk = clk_get(NULL, "adc_clk");
	clk_enable(adc_clk);

	tc0_clk = clk_get(NULL, "tc0_clk");
	clk_enable(tc0_clk);

	mck_clk = clk_get(NULL, "mck");
	MCK = clk_get_rate(mck_clk);

	ret = alloc_chrdev_region(&at91adc_devno, 0, 1, "at91adc");
	if (ret) {
		printk(KERN_ERR "at91adc: Failed to allocate char dev region\n");
		goto exit;
	}

	cdev_init(&at91adc_cdev, &at91adc_fops);
	at91adc_cdev.owner = THIS_MODULE;
	ret = cdev_add(&at91adc_cdev, at91adc_devno, 1);
	if (ret) {
		printk(KERN_ERR "at91adc: Unable register character device\n");
		goto exit1;
	}

	/* Map ADC registers to the current address space */
	adc_base = ioremap_nocache(AT91SAM9260_BASE_ADC, SZ_16K);
	if(!adc_base) {
		printk(KERN_ERR "at91adc: Can't remap ADC register area\n");
		ret = -EACCES;
		goto exit1;
	}

	/* Map TC0 registers to the current address space */
	tc0_base = ioremap_nocache(AT91SAM9260_BASE_TC0, 64);
	if (!tc0_base) {
		printk(KERN_ERR "at91adc: Can't remap TC0 register area\n");
		ret = -EACCES;
		goto exit2;
	}

	/* Allocate buffer memory for storing ADC values */
	adc_buf = kzalloc(MAX_BUF_LEN, GFP_KERNEL);
	if(!adc_buf) {
		printk(KERN_ERR "at91adc: Memory allocation failed\n");
		ret = -ENOMEM;
		goto exit3;
	}

	rx_dma = dma_map_single(NULL, adc_buf, MAX_BUF_LEN, DMA_FROM_DEVICE);
	at91_adc_write(ATMEL_PDC_RPR, rx_dma);
	/* wait trigger condition */
	at91_adc_write(ATMEL_PDC_RCR, 1);
	at91_adc_write(ATMEL_PDC_RNCR, 0);

    for(ADCCLK = 5000000; ADCCLK; ADCCLK--)
		if(MCK == MCK / (ADCCLK * 2) * (ADCCLK * 2))
			break;                                                              
                                                                                
    if((!ADCCLK) || (255 < (prescal = MCK / (ADCCLK * 2) -1))) {
		printk(KERN_ERR "at91adc: Can't scale ADC clock\n");
		ret = -ENOMEM;
		goto exit4;
	}

	/* MUX GPIO for ADC operation */
	at91_set_A_periph(AT91_PIN_PC0, 0);
	/* Reset the ADC */
	at91_adc_write(AT91_ADC_CR, AT91_ADC_SWRST);
	/* Enable ADC channel 0 */
	at91_adc_write(AT91_ADC_CHER, AT91_ADC_CH(0));

	/* Configure ADC mode register */
	shtim = ADCCLK * 24 / 10000000 - 1;   
	/* Enable hardware trigger from TIOA output from TC0, 8bit resolution */
	at91_adc_write(AT91_ADC_MR, AT91_ADC_TRGEN | AT91_ADC_TRGSEL_TC0
	               | AT91_ADC_LOWRES | AT91_ADC_PRESCAL_(prescal)
				   | AT91_ADC_SHTIM_(shtim));

	/* Configure TC0 in waveform mode, MCK/32 */
	at91_tc_write(AT91_TC_CMR, AT91_TC_WAVE | AT91_TC_WAVESEL_UP_AUTO
	            | AT91_TC_TIMER_CLOCK3 | AT91_TC_ACPA_SET | AT91_TC_ACPC_CLEAR);

	curr_srate = srate = align_rate(MAX_SRATE / 2);
	/* set sample rate */
	at91_tc_write(AT91_TC_RC, (MCK/32) / srate);
	/* duty 50 % */
	at91_tc_write(AT91_TC_RA, (MCK/32) / (srate * 2));

	/* set initial data */
	pref = 255;
	hsize = PCF8833;
	state = SINGLE_BYTE;
	threshold = START_THRESHOLD;

//	at91_adc_write(AT91_ADC_IDR, -1);
	/* Install ADC interrupt  */
	ret = request_irq(AT91SAM9260_ID_ADC, at91adc_isr, 0, "at91adc", NULL);
	if (ret) {
		printk(KERN_ERR "at91adc: ADC interrupt request failed\n");
		ret = -EBUSY;
		goto exit4;
	}

	printk(KERN_INFO "at91adc: ADC chanel 0, ADCClock %dHz, Sample & Hold Time %dus,
	                 sampling %dHz, buffer length %d\n", ADCCLK,
	                 (shtim + 1) * 1000000 / ADCCLK, curr_srate, hsize);

	at91_adc_write(ATMEL_PDC_PTCR, ATMEL_PDC_RXTEN);
	at91_adc_write(AT91_ADC_IER, AT91_ADC_ENDRX);
	at91_tc_write(AT91_TC_CCR, AT91_TC_SWTRG | AT91_TC_CLKEN);

	return 0;

exit4:
	kfree(adc_buf);
exit3:
	iounmap(tc0_base);
exit2:
	iounmap(adc_base);
exit1:
	unregister_chrdev_region(at91adc_devno, 1);
exit:
	clk_disable(tc0_clk);
	clk_disable(adc_clk);

	return ret;
}

static void __exit at91adc_exit(void)
{
	clk_disable(adc_clk);
	clk_disable(tc0_clk);

	free_irq(AT91SAM9260_ID_ADC, NULL);

	iounmap(adc_base);
	iounmap(tc0_base);

	kfree(adc_buf);
	/* Free device number */
	unregister_chrdev_region(at91adc_devno, 1);
}

static int at91adc_open(struct inode *inode, struct file *file)
{
	return 0;
}

static int at91adc_release(struct inode *inode, struct file *file)
{
	return 0;
}

static ssize_t at91adc_read(struct file *file,
                            char __iomem *buf,
                            size_t count,
                            loff_t *ppos)
{
	unsigned long p = *ppos;

	if(p >= hsize)
		return 0;
	if(count > hsize)
		count = hsize;
	if(count + p > hsize)
		count = hsize - p;

	if(count) {
		count -= copy_to_user(buf, adc_buf + p, count);
		if(!count)
			return -EFAULT;
		*ppos += count;
	}

	return count;
}

static int at91adc_ioctl(struct inode *inode, struct file *file,
                         unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
	case GETRATE:
		return put_user(curr_srate, (int __user *)arg); 
	case SETRATE:
		{
			int new_srate;

			if(get_user(new_srate, (int __user *)arg))               
				return -EFAULT;                                 
			if((new_srate < MIN_SRATE) || (new_srate > MAX_SRATE))
				return -EINVAL;
			srate = new_srate;
		}
		return 0;
	case GETHSIZE:
		return put_user(h_size, (int __user *)arg); 
	case SETHSIZE:
		{
			int new_hsize;

			if(get_user(new_hsize, (int __user *)arg))               
				return -EFAULT;                                 
			if(new_hsize > MAX_BUF_LEN)
				return -EINVAL;
			hsize = new_hsize;
		}
		return 0;
	case GETTHRESHOLD:
		return put_user(threshold, (int __user *)arg); 
	case SETTHRESHOLD:
		{
			int new_threshold;

			if(get_user(new_threshold, (int __user *)arg))               
				return -EFAULT;                                 
			if(new_threshold > MAX_THRESHOLD)
				return -EINVAL;
			threshold = new_threshold;
		}
		return 0;
	}

	return -EINVAL;
}
						 
struct file_operations at91adc_fops = {
	.owner   = THIS_MODULE,
	.open    = at91adc_open,
	.read    = at91adc_read,
	.ioctl   = at91adc_ioctl,
	.release = at91adc_release,
};

module_init(at91adc_init);
module_exit(at91adc_exit);

MODULE_AUTHOR("Alexander Kudjashev");
MODULE_DESCRIPTION("AT91SAM9260 ADC driver");
MODULE_LICENSE("GPL");
